from typing import Dict, Sequence, Tuple, Union

import jax

from .boxes import box_iou_similarity

jnp = jax.numpy

""" Various functions deals with segmentation pathces
    All functions here takes unbatched input. Use vmap to convert to batched data
"""


def gather_patches(features, locations, patch_size: int) -> tuple:
    """extract feature patches according to a list of locations
    Args:
        features: [H,W,C] standard 2D feature map
        locations: [N, 2] float32, scaled 0..1
        patch_size: int
    Returns:
        patches: [N, patch_size, patch_size, C]
        y0: [N]: y0 coordinates of patches
        x0: [N]: x0 coordinates of patches
    """
    height, width, _ = features.shape

    locations *= jnp.array([height, width])
    # i_locations = (locations + .5).astype(int)
    i_locations = locations.astype(int)
    i_locations_x = jnp.clip(i_locations[:, 1], 0, width - 1)
    i_locations_y = jnp.clip(i_locations[:, 0], 0, height - 1)
    yy, xx = jnp.mgrid[:patch_size, :patch_size] - patch_size // 2
    xx += i_locations_x[:, None, None]
    yy += i_locations_y[:, None, None]
    remainder = locations - jnp.stack([i_locations_y, i_locations_x], axis=-1)

    # padding to avoid out-of-bound
    padding_size = patch_size // 2 + 1
    paddings = [
        [padding_size, padding_size],
        [padding_size, padding_size],
        [0, 0],
    ]
    padded_features = jnp.pad(features, paddings)

    patches = padded_features[yy + padding_size, xx + padding_size, :]

    return patches, yy[:, 0, 0], xx[:, 0, 0], remainder


def _get_patch_data(pred):
    if isinstance(pred, dict):
        patches = pred["instance_output"]
        yc = pred["instance_yc"]
        xc = pred["instance_xc"]
    else:
        patches, yc, xc = pred

    if patches.ndim > yc.ndim:
        patches = patches.squeeze(-1)

    return patches, yc, xc


def bboxes_of_patches(
    pred: Union[Sequence, Dict], threshold: float = 0.5
) -> jnp.ndarray:
    """
    Args:
        pred: either a tuple or a dict containing three arrays:
            instance_outputs: [n, patch_size, patch_size, 1] float
            instance_y_coords: [n, patch_size, patch_size] y patch coords
            instance_x_coords: [n, patch_size, patch_size] x patch coords
        threshold: float
    Returns:
        bboxes: [n, 4] int
        bboox for empty patches are filled with -1
    """
    patches, yy, xx = _get_patch_data(pred)

    _, d0, d1 = patches.shape
    row_mask = jnp.any(patches > threshold, axis=1)
    col_mask = jnp.any(patches > threshold, axis=2)

    min_col = row_mask.argmax(axis=1)
    max_col = d1 - row_mask[:, ::-1].argmax(axis=1)

    min_row = col_mask.argmax(axis=1)
    max_row = d0 - col_mask[:, ::-1].argmax(axis=1)

    min_row += yy[:, 0, 0]
    max_row += yy[:, 0, 0]
    min_col += xx[:, 0, 0]
    max_col += xx[:, 0, 0]

    is_valid = row_mask.any(axis=1, keepdims=True)
    bboxes = jnp.stack([min_row, min_col, max_row, max_col], axis=-1)
    bboxes = jnp.where(is_valid, bboxes, -1)

    return bboxes


def indices_of_patches(
    pred: Union[Sequence, Dict],
    input_size: Tuple[int, int] = None,
    threshold: float = 0.5,
) -> tuple:
    """
    Args:
        pred: model output (unbatched):
        threshold: float
    Returns:
        data for MaskIndices
    """
    patches, yy, xx = _get_patch_data(pred)

    indices = patches >= threshold
    yy = yy[indices]
    xx = xx[indices]
    rowids = jnp.argwhere(indices)[:, 0]

    if input_size is not None:
        height, width = input_size
        valid_coords = ((yy >= 0) & (yy < height)) & ((xx >= 0) & (xx < width))
        yy = yy[valid_coords]
        xx = xx[valid_coords]
        rowids = rowids[valid_coords]

    return jnp.stack([yy, xx, rowids], axis=-1)


def iou_patches_and_labels(pred, labels, BLOCK_SIZE=128):
    """Compute iou between two sets of segmentations.
    The first set is defined by patches; the second by image labels
    Args:
        pred: model output (unbatched):
        threshold: float
    Returns:
        [n, m] iou values.
    """
    patches, yc, xc = _get_patch_data(pred)
    patches = patches >= 0.5
    pred_areas = jnp.count_nonzero(patches, axis=(-1, -2))

    # this rely on JAX's out-of-bound indexing behavior
    padded_labels = jnp.pad(labels, [[1, 1], [1, 1]], constant_values=-1)
    gt_patches = padded_labels[yc + 1, xc + 1]
    max_indices = ((labels.max() - 1) // BLOCK_SIZE + 1) * BLOCK_SIZE
    all_gt_areas = jnp.count_nonzero(
        labels[:, :, None] == jnp.arange(max_indices) + 1, axis=(0, 1)
    )

    ious = []
    # FIXME change to scan
    for k in range(1, labels.max() + 1, BLOCK_SIZE):
        gt_p = gt_patches == jnp.arange(k, k + BLOCK_SIZE).reshape(
            -1, 1, 1, 1
        )  # [B, N, s, s]
        gt_areas = all_gt_areas[k - 1 : k + BLOCK_SIZE - 1]
        intersect = jnp.count_nonzero(gt_p & patches, axis=(-1, -2))  # [B, N]
        ious.append(intersect / (pred_areas + gt_areas[:, None] - intersect + 1.0e-8))

    ious = jnp.concatenate(ious, axis=0).transpose()  # [N, B * b]
    ious = ious[:, : labels.max()]
    return ious


def patches_to_segmentations(pred, input_size, threshold=0.5):
    """expand patches to the full image size
    Args:
        pred: model output (unbatched):
        input_size: tuple(int, int)
        threshold: float
    Returns:
        segmentations: [n, height, width]
    """
    patches, yc, xc = _get_patch_data(pred)
    n_patches, patch_size, _ = yc.shape

    page_nums = jnp.arange(n_patches)
    segms = jnp.zeros((n_patches,) + input_size)
    segms = segms.at[page_nums[:, None, None], yc, xc].set(patches)

    return (segms >= threshold).astype(int)


def patches_to_label(pred, input_size, score_threshold=0.5, threshold=0.5):
    """convert patches to the image label
    Args:
        pred: model output (unbatched):
        threshold: float
    Returns:
        label: [height, width]
    """
    label = jnp.zeros(input_size)

    if "training_locations" in pred:
        mask = pred["instance_mask"]
    else:
        mask = pred["pred_scores"] >= score_threshold

    n = jnp.count_nonzero(mask)
    pr = pred["instance_output"]
    seq = jnp.arange(pr.shape[0])
    seq = n - jnp.where(seq < n, seq, n)
    pr = (pr > threshold) * (seq[:, None, None])
    yc = pred["instance_yc"]
    xc = pred["instance_xc"]

    label = label.at[yc, xc].max(pr)

    return label



# copied from https://github.com/cgarciae/simple-pytree

import dataclasses
import typing as tp
from abc import ABCMeta
from copy import copy

import jax

P = tp.TypeVar("P", bound="Pytree")


def field(
    default: tp.Any = dataclasses.MISSING,
    *,
    pytree_node: bool = True,
    default_factory: tp.Any = dataclasses.MISSING,
    init: bool = True,
    repr: bool = True,
    hash: tp.Optional[bool] = None,
    compare: bool = True,
    metadata: tp.Optional[tp.Mapping[str, tp.Any]] = None,
):
    if metadata is None:
        metadata = {}
    else:
        metadata = dict(metadata)

    if "pytree_node" in metadata:
        raise ValueError("'pytree_node' found in metadata")

    metadata["pytree_node"] = pytree_node

    return dataclasses.field(
        default=default,
        default_factory=default_factory,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
        metadata=metadata,
    )


def static_field(
    default: tp.Any = dataclasses.MISSING,
    *,
    default_factory: tp.Any = dataclasses.MISSING,
    init: bool = True,
    repr: bool = True,
    hash: tp.Optional[bool] = None,
    compare: bool = True,
    metadata: tp.Optional[tp.Mapping[str, tp.Any]] = None,
):
    return field(
        default=default,
        pytree_node=False,
        default_factory=default_factory,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
        metadata=metadata,
    )


class PytreeMeta(ABCMeta):
    def __call__(self: tp.Type[P], *args: tp.Any, **kwds: tp.Any) -> P:
        obj: Pytree = super().__call__(*args, **kwds)
        object.__setattr__(obj, "_pytree__initialized", True)
        return obj


class Pytree(metaclass=PytreeMeta):
    _pytree__initialized: bool
    _pytree__static_fields: tp.Set[str]
    _pytree__class_is_mutable: bool

    def __init_subclass__(cls, mutable: bool = False):
        jax.tree_util.register_pytree_node(
            cls,
            flatten_func=tree_flatten,
            unflatten_func=lambda *_args: tree_unflatten(cls, *_args),
        )

        # init class variables
        cls._pytree__initialized = False  # initialize mutable
        cls._pytree__static_fields = set()
        cls._pytree__class_is_mutable = mutable

        # get class info
        class_vars = _get_all_class_vars(cls)

        for field, value in class_vars.items():
            if "_pytree__" in field or (
                isinstance(value, dataclasses.Field)
                and value.metadata is not None
                and not value.metadata.get("pytree_node", True)
            ):
                cls._pytree__static_fields.add(field)

    def replace(self: P, **kwargs: tp.Any) -> P:
        """
        Replace the values of the fields of the object with the values of the
        keyword arguments. If the object is a dataclass, `dataclasses.replace`
        will be used. Otherwise, a new object will be created with the same
        type as the original object.
        """
        if dataclasses.is_dataclass(self):
            return dataclasses.replace(self, **kwargs)

        fields = vars(self)
        for key in kwargs:
            if key not in fields:
                raise ValueError(f"'{key}' is not a field of {type(self).__name__}")

        pytree = copy(self)
        pytree.__dict__.update(kwargs)
        return pytree

    if not tp.TYPE_CHECKING:

        def __setattr__(self: P, field: str, value: tp.Any):
            if self._pytree__initialized and not self._pytree__class_is_mutable:
                raise AttributeError(
                    f"{type(self)} is immutable, trying to update field {field}"
                )

            object.__setattr__(self, field, value)


def tree_flatten(pytree: Pytree):
    node_fields = {}
    static_fields = {}

    for field, value in vars(pytree).items():
        if field in pytree._pytree__static_fields:
            static_fields[field] = value
        else:
            node_fields[field] = value

    return (node_fields,), static_fields


def tree_unflatten(cls: tp.Type[P], static_fields, children):
    (node_fields,) = children
    pytree = cls.__new__(cls)
    pytree.__dict__.update(node_fields, **static_fields)
    return pytree


def _get_all_class_vars(cls: type) -> tp.Dict[str, tp.Any]:
    d = {}
    for c in reversed(cls.mro()):
        if hasattr(c, "__dict__"):
            d.update(vars(c))
    return d



import inspect
import typing as tp
from dataclasses import asdict

import flax.linen as nn
import jax
import jax.numpy as jnp
import jax.tree_util as jtu
import optax
from flax.core.frozen_dict import freeze, unfreeze
from optax import GradientTransformation

from .loss import Loss
from .pytree import Pytree, static_field


class WrappedModule(Pytree, mutable=True):
    _variables: dict
    _flax_module: nn.Module = static_field()
    initialized: bool = static_field()

    def __init__(self, obj):
        if not isinstance(obj, nn.Module):
            raise ValueError("Can only wrap flax modules")

        self._flax_module = obj
        self._variables = {}
        self.initialized = False
        # self.__doc__ = obj.__doc__
        # self.__annotations__.update(obj.__annotations__)

    def init(self, *args, **kwargs):
        _ = self.init_with_output(*args, **kwargs)

    def init_with_output(self, *args, **kwargs):
        output, self._variables = self._flax_module.init_with_output(*args, **kwargs)
        self.initialized = True
        return output

    def update_variables(self, vars):
        old_vars = unfreeze(self._variables)
        old_vars.update(vars)
        self._variables = freeze(old_vars)

    def get_variables(self, col="params"):
        return self._variables[col]

    def __str__(self, *args, **kwargs):
        return f"Pytree({self._flax_module.__str__(*args, **kwargs)}))"

    def __repr__(self, *args, **kwargs):
        return f"Pytree({self._flax_module.__repr__(*args, **kwargs)})"

    def __call__(self, *args, **kwargs):
        if "mutable" in kwargs:
            raise ValueError('expected kwargs "mutable".')
        output, self._variables = self._flax_module.apply(
            self._variables, *args, mutable=True, **kwargs
        )
        return output

    def __getattr__(self, attr):
        if attr == "apply":
            raise AttributeError(
                "Do not call apply() in a pytree-ized moduel. Use __call__() instead"
            )
        if attr == "vars":
            return self._variables

        return getattr((self._flax_module), attr)

    def get_config(self):
        return asdict(self._flax_module)

    # def parameter_flags(self):
    #     all_nodes = jtu.tree_map(lambda x: x, self)
    #     all_nodes._variables = unfreeze(all_nodes._variables)
    #     params_ids = jtu.tree_leaves(jtu.tree_map(id, all_nodes._variables['params']))
    #     flags = jtu.tree_map(lambda x: id(x) in params_ids, all_nodes)
    #     flags._variables = freeze(flags._variables)
    #     return flags


class WrappedGT(Pytree, mutable=True):
    _states: tp.Any
    # _tx_core: GradientTransformation = static_field()
    _tx: GradientTransformation = static_field()
    initialized: bool = static_field()

    def __init__(self, obj):
        if not isinstance(obj, GradientTransformation):
            raise ValueError("Can only wrap optax optimizes")
        self._tx = obj
        self._states = None
        self.initialized = False

    def init(self, tree):
        # partitions = {True: self._tx_core, False: optax.set_to_zero()}
        # flags = jtu.tree_leaves(tree.parameter_flags())
        # self._tx = optax.multi_transform(partitions, flags)
        # self._state = self._tx.init(jtu.tree_leaves(tree))
        self._state = self._tx.init(jtu.tree_leaves(tree))
        self.initialized = True

    def update(self, grad, tree):
        tree_lfs, st1 = jtu.tree_flatten(tree)
        grad_lfs, st2 = jtu.tree_flatten(grad)
        assert st1 == st2
        updates, self._state = self._tx.update(grad_lfs, self._state, tree_lfs)
        tree_lfs = optax.apply_updates(tree_lfs, updates)
        tree = jtu.tree_unflatten(st1, tree_lfs)
        return tree


class LossLog(Pytree, mutable=True):
    _losses: tp.Sequence[Loss] = static_field()

    def __init__(self, losses):
        if isinstance(losses, Loss):
            self._losses = [losses]
        else:
            self._losses = losses

        self.reset()

    def reset(self):
        self._cnts = {}
        self._sums = {}
        for loss in self._losses:
            self._cnts.update({loss.name: jnp.asarray(0.0)})
            self._sums.update({loss.name: jnp.asarray(0.0)})

    def update(self, **kwargs):
        loss_log = {loss_fn.name: loss_fn(**kwargs) for loss_fn in self._losses}

        for k, v in loss_log.items():
            self._cnts[k] += 1
            self._sums[k] += v

        return sum(loss_log.values()), loss_log

    def compute(self):
        results = {k: self._sums[k] / self._cnts[k] for k in self._cnts}
        return results



import dataclasses
from functools import partial

import jax
import jax.numpy as jnp
import numpy as np


def sample_unselected(selected, p, p_miss, *, key):
    @jax.vmap
    def _mapped_choice(keys, weights):
        n_target = weights.size
        return jax.random.choice(
            keys,
            n_target,
            shape=[
                1,
            ],
            p=weights,
        )

    n_sample, n_target = selected.shape
    total_mass = p_miss + p.sum()
    p_miss = p_miss / total_mass
    p = p / total_mass
    w = p * (1 - selected)  # [n_samples, n_targets]
    w = jnp.pad(w, [[0, 0], [0, 1]], constant_values=p_miss)
    w = w / w.sum(axis=-1, keepdims=True)  # noramlze each row
    assert w.shape == (n_sample, n_target + 1)

    keys = jax.random.split(key, n_sample)
    selection = _mapped_choice(keys, w)

    # formatting issue
    selection = jnp.where(selection == n_target, -1, selection)

    return selection


def update_history(history, selected, selection, k):
    n_sample = history.shape[0]
    history = history.at[:, k].set(selection[:, 0])
    # history = jnp.concatenate([history, selection], axis=-1)
    selected = jnp.pad(
        selected, [[0, 0], [0, 1]]
    )  # pad an extra column to catch both -1
    selected = selected.at[jnp.arange(n_sample), selection[:, 0]].set(1)
    selected = selected[:, :-1]  # remove padding

    return history, selected


def weight_history(selected, p, p_miss):
    total_mass = p_miss + p.sum()
    p_miss = p_miss / total_mass
    p = p / total_mass

    w = (p * (1 - selected)).sum(axis=-1) + p_miss
    return w


def resample_history(history, selected, p, p_miss, *, key):
    n_sample = selected.shape[0]
    w = weight_history(selected, p, p_miss)
    w = w / w.sum()
    # print(w.shape)
    c = jax.random.choice(key, n_sample, p=w, shape=[n_sample])
    history = history[c]
    selected = selected[c]

    return history, selected


def seq_choice(wt, w_miss, prev=None, *, key, n_sample=256, resampling=True):
    """
    Args
        wt: ndarray [M x N]
        w_miss: zero or 1d array, if 1d size is [M],
        prev: when sampling division, prev is not None, and w_miss represent weight for no-division
        n_samples: int optional
        resampling: if set to Flase, the function do not resample, but return samples weights instead
    returns: a single sample
        history: [n_sample, M]
        selected: [n_sample, N]
    """

    w_miss = jnp.asarray(w_miss)
    w_miss = w_miss.reshape(w_miss.size)

    n_source, n_target = wt.shape
    log_weight = jnp.zeros([n_sample])

    if prev is None:
        history = jnp.zeros(shape=[n_sample, n_source], dtype=int)
        selected = jnp.zeros(shape=[n_sample, n_target], dtype=int)
    elif resampling:  # for sampling cell div, a selection is provided
        raise ValueError("cannot resample if supplied with previous results")
    else:
        selected = prev
        history = jnp.zeros(shape=[n_sample, n_source], dtype=int)
        log_weight = jnp.log(weight_history(selected, wt[0], w_miss[0]))

    def _for_inner(k, state):
        history, selected, key, log_weight = state
        key, sample_key, resample_key = jax.random.split(key, 3)

        # sample current availabel position
        selection = sample_unselected(selected, wt[k], w_miss[k], key=sample_key)

        # update history
        history, selected = update_history(history, selected, selection, k)

        # resample based on next row
        if resampling:
            history, selected = resample_history(
                history, selected, wt[k + 1], w_miss[k + 1], key=resample_key
            )
        else:
            log_weight += jnp.log(weight_history(selected, wt[k + 1], w_miss[k + 1]))

        return history, selected, key, log_weight

    history, selected, key, log_weight = jax.lax.fori_loop(
        0,
        n_source - 1,
        lambda k, state: jax.lax.cond(
            jnp.any(wt[k] > 0),
            _for_inner,
            lambda _, state: state,
            k,
            state,
        ),
        (history, selected, key, log_weight),
    )

    # final iter
    k = n_source - 1
    selection = sample_unselected(selected, wt[k], w_miss[k], key=key)
    history, selected = update_history(history, selected, selection, k)

    if resampling:
        return history, selected
    else:
        return history, selected, log_weight


def sample_step(key, wts, w_miss, w_nd, n_sub_sample=256):
    """
    Args:
        key: a single key
        wts: [n_samples, n_source, n_target], pad with 0 will ensure all invalid rows return -1
        w_miss: float constant
        w_nd: [n_samples, n_source]
    Returns:
        history: [n_samples, n_source], -1 means missing
        history_div: [n_samples, n_source], -1 means no division
        selected: [n_samples, n_target], 1/0
    """
    n_sample, n_source, n_target = wts.shape

    # PAD_BLOCK = 32

    # padto = (max(n_source, n_target) - 1) // PAD_BLOCK * PAD_BLOCK + PAD_BLOCK
    # padded_wts = jnp.pad(wts, [[0, 0], [0, padto - n_source], [0, padto - n_target]])
    # padded_nd = jnp.pad(w_nd, [[0, 0], [0, padto - n_source]], constant_values=1.0)

    w_miss = jnp.asarray(w_miss).repeat(n_sample).reshape([n_sample, 1])

    def _inner(key, wts, w_miss, w_nd):
        key, k1, k2, k3 = jax.random.split(key, 4)
        k1 = jax.random.split(k1, n_sample)
        k2 = jax.random.split(k2, n_sample)
        k3 = jax.random.split(k3, n_sample)

        history, selected = jax.vmap(partial(seq_choice, n_sample=n_sub_sample))(
            wts,
            w_miss,
            key=k1,
        )

        w_nd = w_nd * (wts.sum(axis=-1) + w_miss) + w_miss
        history_div, selected, weights = jax.vmap(
            partial(seq_choice, n_sample=n_sub_sample, resampling=False)
        )(wts, w_nd, selected, key=k2)
        weights = jnp.exp(weights)
        weights /= weights.sum(axis=-1, keepdims=True)
        rs = jax.vmap(partial(jax.random.choice, a=weights.shape[1]))(k3, p=weights)

        history = history[jnp.arange(n_sample), rs]  # just need one sample
        history_div = history_div[jnp.arange(n_sample), rs]
        selected = selected[jnp.arange(n_sample), rs]

        reversed = (history == -1) & (history_div != -1)
        tmp_history = jnp.where(reversed, history_div, history)
        history_div = jnp.where(reversed, history, history_div)
        history = tmp_history

        return history, history_div, selected

    # history, history_div, selected = jax.jit(_inner)(key, padded_wts, w_miss, padded_nd)
    # history = history[:, :n_source]
    # history_div = history_div[:, :n_source]
    # selected = selected[:, :n_target]

    history, history_div, selected = _inner(key, wts, w_miss, w_nd)

    return history, history_div, selected


def post_process_step(history):
    selected0 = history["samples"][-2]["selected"]
    selected1 = history["samples"][-1]["selected"]
    link = history["samples"][-2]["history"]
    div = history["samples"][-2]["history_div"]
    lifetime0 = history["samples"][-2]["lifetime"]
    yx1 = history["detections"][-1]["yx"]
    yx0 = history["detections"][-2]["yx"]
    id1 = history["detections"][-1]["id"]
    id0 = history["detections"][-2]["id"]

    history["detections"][-2]["yx_next"] = yx1
    history["detections"][-2]["id_next"] = id1

    n_target = yx1.shape[0]

    def _update_lifetime(lifetime0, link, div):
        lifetime = jnp.zeros([n_target + 1])
        link_no_div = jnp.where(div != -1, -1, link)
        lifetime = lifetime.at[link_no_div].set(lifetime0 + 1)[:-1]
        lifetime_extra = jnp.where(link == -1, lifetime0 + 1, 0)
        return jnp.concatenate([lifetime, lifetime_extra])

    indicator_for_yx0 = (selected0 == 1) & (link == -1)  # [n_samples, n_source]
    indicator_for_yx1 = selected1 == 1  # [n_samples, n_target]
    indicator_all = np.concatenate(
        [indicator_for_yx1, indicator_for_yx0], axis=-1
    )  # [n_sample, n_source+n+target]
    is_not_empty = indicator_all.any(axis=0)  # [n_source+n_target]

    yx_all = np.concatenate([yx1, yx0])  # [n_sourc+n_target, 2]
    id_all = np.concatenate([id1, id0])
    lifetime1 = np.asarray(jax.vmap(_update_lifetime)(lifetime0, link, div))
    # print(lifetime1.shape)

    history["detections"][-1].update(
        dict(
            yx=yx_all[is_not_empty],
            id=id_all[is_not_empty],
        )
    )
    history["samples"][-1].update(
        dict(
            selected=indicator_all[:, is_not_empty],
            lifetime=lifetime1[:, is_not_empty],
        )
    )

    return history


def get_tracking_weights(yx0, yx1, gamma):
    delta = ((yx1[None, :, :] - yx0[:, None, :]) ** 2).sum(
        axis=-1
    )  # [n_source, n_target]
    wts = jnp.exp(-delta / 2 * gamma * gamma)
    return wts


def compute_div_p(lifetime, avg, scale, limit):
    p0 = (1 / scale) ** avg
    return p0 * (scale**lifetime) + limit


@dataclasses.dataclass
class HyperParams:
    gamma: float = 0.1
    div_avg: float = 10.0
    div_limit: float = 0.1
    div_scale: float = 0.5
    w_miss: float = 0.01
    logit_scale: float = 1.0
    logit_offset: float = 0.0
    missing_logit: float = -0.5
    n_sub_sample: int = 64


def track_to_next_frame(key, history, nextframe, hyper_params):
    """
    Args:
        key: rng
        nextframe: dict
          yx: [n_target, 2], yx locations
          id: [n_target] int
          logit: [n_target], logit of each yx1 location
        history: dict['yx': [array,...], 'samples': [{'selected':array, 'history':array, 'history_div':array, 'lifetime':array},...]]
            yx: [n_source, 2], yx locations from frame 0
            selected: [n_samples, n_source]: indicators of source locations that are being tracked
            lifetime: [n_samples, n_source] how long has each sample been without division
        haper_params: HyperParam dataclass
    """
    yx0 = history["detections"][-1]["yx"]
    selected0 = history["samples"][-1]["selected"]
    lifetime = history["samples"][-1]["lifetime"]
    n_sample, _ = selected0.shape

    yx1 = np.asarray(nextframe["yx"])
    id1 = np.asarray(nextframe["id"])
    logit = np.asarray(nextframe["logit"])

    wts = get_tracking_weights(yx0, yx1, hyper_params.gamma)  # [n_source, n_target]
    wts = jnp.where(
        selected0[:, :, None],
        wts[
            None,
            :,
            :,
        ],
        0.0,
    )  # masking out unselected source locations

    w_nd = compute_div_p(
        lifetime,
        hyper_params.div_avg,
        hyper_params.div_scale,
        hyper_params.div_limit,
    )
    new_history, new_history_div, selected1 = sample_step(
        key, wts, hyper_params.w_miss, w_nd, n_sub_sample=hyper_params.n_sub_sample
    )

    history["detections"].append(
        dict(
            yx=yx1,
            id=id1,
            logit=logit,
        )
    )
    history["samples"][-1].update(
        dict(
            history=np.asarray(new_history),
            history_div=np.asarray(new_history_div),
        )
    )
    history["samples"].append(
        dict(
            selected=np.asarray(selected1),
        )
    )

    target_logit = logit * hyper_params.logit_scale + hyper_params.logit_offset
    target_logit = (target_logit * selected1).sum(axis=-1)
    is_missing = (selected0 == 1) & (new_history == -1)
    target_logit += np.count_nonzero(is_missing, axis=1) * hyper_params.missing_logit
    weights = jax.nn.softmax(target_logit)
    rs = np.asarray(jax.random.choice(key, n_sample, [n_sample], p=weights))

    history["samples"] = jax.tree_map(lambda v: v[rs], history["samples"])

    history = post_process_step(history)

    return history



import re
import typing as tp

InputLike = tp.Union[tp.Any, tp.Tuple[tp.Any, ...], tp.Dict[str, tp.Any], "Inputs"]


def _unique_name(
    names: tp.Set[str],
    name: str,
):

    if name in names:

        match = re.match(r"(.*?)(\d*)$", name)
        assert match is not None

        name = match[1]
        num_part = match[2]

        i = int(num_part) if num_part else 2
        str_template = f"{{name}}{{i:0{len(num_part)}}}"

        while str_template.format(name=name, i=i) in names:
            i += 1

        name = str_template.format(name=name, i=i)

    names.add(name)
    return name


def _unique_names(
    names: tp.Iterable[str],
    *,
    existing_names: tp.Optional[tp.Set[str]] = None,
) -> tp.Iterable[str]:
    if existing_names is None:
        existing_names = set()

    for name in names:
        yield _unique_name(existing_names, name)


def _lower_snake_case(s: str) -> str:
    s = re.sub(r"(?<!^)(?=[A-Z])", "_", s).lower()
    parts = s.split("_")
    output_parts = []

    for i in range(len(parts)):
        if i == 0 or len(parts[i - 1]) > 1:
            output_parts.append(parts[i])
        else:
            output_parts[-1] += parts[i]

    return "_".join(output_parts)


def _get_name(obj) -> str:
    if hasattr(obj, "name") and obj.name:
        return obj.name
    elif hasattr(obj, "__name__") and obj.__name__:
        return _lower_snake_case(obj.__name__)
    elif hasattr(obj, "__class__") and obj.__class__.__name__:
        return _lower_snake_case(obj.__class__.__name__)
    else:
        raise ValueError(f"Could not get name for: {obj}")


class Inputs:
    args: tp.Tuple[tp.Any, ...]
    kwargs: tp.Dict[str, tp.Any]

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    @classmethod
    def from_value(cls, value: InputLike) -> "Inputs":
        if isinstance(value, cls):
            return value
        elif isinstance(value, tuple):
            return cls(*value)
        elif isinstance(value, dict):
            return cls(**value)
        else:
            return cls(value)
